@@ -52,7 +52,6 @@
 #include <libgd/gd-notification.h>
 
 #define GNOME_DESKTOP_INPUT_SOURCES_DIR "org.gnome.desktop.input-sources"
-#define KEY_CURRENT_INPUT_SOURCE "current"
 #define KEY_INPUT_SOURCES        "sources"
 
 #define GNOME_SYSTEM_LOCALE_DIR "org.gnome.system.locale"
@@ -172,7 +171,8 @@ cc_region_panel_finalize (GObject *object)
         g_free (priv->system_region);
 
         chooser = g_object_get_data (G_OBJECT (self), "input-chooser");
-        gtk_widget_destroy (chooser);
+        if (chooser)
+                gtk_widget_destroy (chooser);
 
 	G_OBJECT_CLASS (cc_region_panel_parent_class)->finalize (object);
 }
@@ -454,31 +454,59 @@ format_response (GtkDialog *chooser,
         gtk_widget_destroy (GTK_WIDGET (chooser));
 }
 
+static const gchar *
+get_effective_language (CcRegionPanel *self)
+{
+        CcRegionPanelPrivate *priv = self->priv;
+        if (priv->login)
+                return priv->system_language;
+        else
+                return priv->language;
+}
+
 static void
-show_language_chooser (CcRegionPanel *self,
-                       const gchar   *language)
+show_language_chooser (CcRegionPanel *self)
 {
         GtkWidget *toplevel;
         GtkWidget *chooser;
 
         toplevel = gtk_widget_get_toplevel (GTK_WIDGET (self));
         chooser = cc_language_chooser_new (toplevel);
-        cc_language_chooser_set_language (chooser, language);
+        cc_language_chooser_set_language (chooser, get_effective_language (self));
         g_signal_connect (chooser, "response",
                           G_CALLBACK (language_response), self);
         gtk_window_present (GTK_WINDOW (chooser));
 }
 
+static const gchar *
+get_effective_region (CcRegionPanel *self)
+{
+        CcRegionPanelPrivate *priv = self->priv;
+        const gchar *region;
+
+        if (priv->login)
+                region = priv->system_region;
+        else
+                region = priv->region;
+
+        /* Region setting might be empty - show the language because
+         * that's what LC_TIME and others will effectively be when the
+         * user logs in again. */
+        if (region == NULL || region[0] == '\0')
+                region = get_effective_language (self);
+
+        return region;
+}
+
 static void
-show_region_chooser (CcRegionPanel *self,
-                     const gchar   *region)
+show_region_chooser (CcRegionPanel *self)
 {
         GtkWidget *toplevel;
         GtkWidget *chooser;
 
         toplevel = gtk_widget_get_toplevel (GTK_WIDGET (self));
         chooser = cc_format_chooser_new (toplevel);
-        cc_format_chooser_set_region (chooser, region);
+        cc_format_chooser_set_region (chooser, get_effective_region (self));
         g_signal_connect (chooser, "response",
                           G_CALLBACK (format_response), self);
         gtk_window_present (GTK_WINDOW (chooser));
@@ -510,10 +538,10 @@ permission_acquired (GObject      *source,
         if (allowed) {
                 switch (priv->op) {
                 case CHOOSE_LANGUAGE:
-                        show_language_chooser (self, priv->system_language);
+                        show_language_chooser (self);
                         break;
                 case CHOOSE_REGION:
-                        show_region_chooser (self, priv->system_region);
+                        show_region_chooser (self);
                         break;
                 case ADD_INPUT:
                         show_input_chooser (self);
@@ -539,10 +567,8 @@ activate_language_row (CcRegionPanel *self,
 	CcRegionPanelPrivate *priv = self->priv;
 
         if (row == priv->language_row) {
-                if (!priv->login) {
-                        show_language_chooser (self, priv->language);
-                } else if (g_permission_get_allowed (priv->permission)) {
-                        show_language_chooser (self, priv->system_language);
+                if (!priv->login || g_permission_get_allowed (priv->permission)) {
+                        show_language_chooser (self);
                 } else if (g_permission_get_can_acquire (priv->permission)) {
                         priv->op = CHOOSE_LANGUAGE;
                         g_permission_acquire_async (priv->permission,
@@ -551,10 +577,8 @@ activate_language_row (CcRegionPanel *self,
                                                     self);
                 }
         } else if (row == priv->formats_row) {
-                if (!priv->login) {
-                        show_region_chooser (self, priv->region);
-                } else if (g_permission_get_allowed (priv->permission)) {
-                        show_region_chooser (self, priv->system_region);
+                if (!priv->login || g_permission_get_allowed (priv->permission)) {
+                        show_region_chooser (self);
                 } else if (g_permission_get_can_acquire (priv->permission)) {
                         priv->op = CHOOSE_REGION;
                         g_permission_acquire_async (priv->permission,
@@ -569,20 +593,9 @@ static void
 update_region_label (CcRegionPanel *self)
 {
         CcRegionPanelPrivate *priv = self->priv;
-        const gchar *region;
+        const gchar *region = get_effective_region (self);
         gchar *name = NULL;
 
-        if (priv->login)
-                region = priv->system_region;
-        else
-                region = priv->region;
-
-        /* Region setting might be empty - show the language because
-         * that's what LC_TIME and others will effectively be when the
-         * user logs in again. */
-        if (region == NULL || region[0] == '\0')
-                region = priv->language;
-
         if (region)
                 name = gnome_get_country_from_locale (region, region);
 
@@ -607,14 +620,9 @@ static void
 update_language_label (CcRegionPanel *self)
 {
 	CcRegionPanelPrivate *priv = self->priv;
-        const gchar *language;
+        const gchar *language = get_effective_language (self);
         gchar *name = NULL;
 
-        if (priv->login)
-                language = priv->system_language;
-        else
-                language = priv->language;
-
         if (language)
                 name = gnome_get_language_from_locale (language, language);
 
@@ -734,19 +742,22 @@ fetch_ibus_engines_result (GObject       *object,
                            GAsyncResult  *result,
                            CcRegionPanel *self)
 {
-        CcRegionPanelPrivate *priv = self->priv;
+        CcRegionPanelPrivate *priv;
         GList *list, *l;
         GError *error;
 
         error = NULL;
-        list = ibus_bus_list_engines_async_finish (priv->ibus, result, &error);
-        g_clear_object (&priv->ibus_cancellable);
+        list = ibus_bus_list_engines_async_finish (IBUS_BUS (object), result, &error);
         if (!list && error) {
-                g_warning ("Couldn't finish IBus request: %s", error->message);
+                if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+                        g_warning ("Couldn't finish IBus request: %s", error->message);
                 g_error_free (error);
                 return;
         }
 
+        priv = self->priv;
+        g_clear_object (&priv->ibus_cancellable);
+
         /* Maps engine ids to engine description objects */
         priv->ibus_engines = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_object_unref);
 
@@ -854,8 +865,8 @@ add_input_row (CcRegionPanel   *self,
         gtk_widget_set_halign (label, GTK_ALIGN_START);
         gtk_widget_set_margin_start (label, 20);
         gtk_widget_set_margin_end (label, 20);
-        gtk_widget_set_margin_top (label, 6);
-        gtk_widget_set_margin_bottom (label, 6);
+        gtk_widget_set_margin_top (label, 12);
+        gtk_widget_set_margin_bottom (label, 12);
         gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 0);
 
         if (strcmp (type, INPUT_SOURCE_TYPE_IBUS) == 0) {
@@ -1049,46 +1060,19 @@ set_input_settings (CcRegionPanel *self)
         const gchar *type;
         const gchar *id;
         GVariantBuilder builder;
-        GVariant *old_sources;
-        const gchar *old_current_type;
-        const gchar *old_current_id;
-        guint old_current;
-        guint old_n_sources;
-        guint index;
         GList *list, *l;
 
-        old_sources = g_settings_get_value (priv->input_settings, KEY_INPUT_SOURCES);
-        old_current = g_settings_get_uint (priv->input_settings, KEY_CURRENT_INPUT_SOURCE);
-        old_n_sources = g_variant_n_children (old_sources);
-
-        if (old_n_sources > 0 && old_current < old_n_sources) {
-                g_variant_get_child (old_sources, old_current,
-                                     "(&s&s)", &old_current_type, &old_current_id);
-        } else {
-                old_current_type = "";
-                old_current_id = "";
-        }
-
         g_variant_builder_init (&builder, G_VARIANT_TYPE ("a(ss)"));
-        index = 0;
         list = gtk_container_get_children (GTK_CONTAINER (priv->input_list));
         for (l = list; l; l = l->next) {
                 type = (const gchar *)g_object_get_data (G_OBJECT (l->data), "type");
                 id = (const gchar *)g_object_get_data (G_OBJECT (l->data), "id");
-                if (index != old_current &&
-                    g_str_equal (type, old_current_type) &&
-                    g_str_equal (id, old_current_id)) {
-                        g_settings_set_uint (priv->input_settings, KEY_CURRENT_INPUT_SOURCE, index);
-                }
                 g_variant_builder_add (&builder, "(ss)", type, id);
-                index += 1;
         }
         g_list_free (list);
 
         g_settings_set_value (priv->input_settings, KEY_INPUT_SOURCES, g_variant_builder_end (&builder));
         g_settings_apply (priv->input_settings);
-
-        g_variant_unref (old_sources);
 }
 
 static void set_localed_input (CcRegionPanel *self);
@@ -1510,8 +1494,8 @@ setup_input_section (CcRegionPanel *self)
         gtk_list_box_set_header_func (GTK_LIST_BOX (priv->input_list),
                                       cc_list_box_update_header_func,
                                       NULL, NULL);
-        g_signal_connect_swapped (priv->input_list, "row-selected",
-                                  G_CALLBACK (update_buttons), self);
+        g_signal_connect_object (priv->input_list, "row-selected",
+                                 G_CALLBACK (update_buttons), self, G_CONNECT_SWAPPED);
 
         g_signal_connect (priv->input_settings, "changed::" KEY_INPUT_SOURCES,
                           G_CALLBACK (input_sources_changed), self);
@@ -1554,9 +1538,6 @@ on_localed_properties_changed (GDBusProxy     *proxy,
                 if (!messages) {
                         messages = lang;
                 }
-                if (!time) {
-                        time = lang;
-                }
                 g_free (priv->system_language);
                 priv->system_language = g_strdup (messages);
                 g_free (priv->system_region);
@@ -1637,7 +1618,8 @@ set_localed_locale (CcRegionPanel *self)
         g_variant_builder_add (b, "s", s);
         g_free (s);
 
-        if (g_strcmp0 (priv->system_language, priv->system_region) != 0) {
+        if (priv->system_region != NULL &&
+            g_strcmp0 (priv->system_language, priv->system_region) != 0) {
                 s = g_strconcat ("LC_TIME=", priv->system_region, NULL);
                 g_variant_builder_add (b, "s", s);
                 g_free (s);
@@ -1810,7 +1792,7 @@ setup_login_button (CcRegionPanel *self)
         g_object_unref (bus);
 
         priv->login_label = WID ("login-label");
-        priv->login_button = gtk_toggle_button_new_with_label (_("Login Screen"));
+        priv->login_button = gtk_toggle_button_new_with_mnemonic (_("Login _Screen"));
         gtk_style_context_add_class (gtk_widget_get_style_context (priv->login_button),
                                      "text-button");
         gtk_widget_set_valign (priv->login_button, GTK_ALIGN_CENTER);
